# -*- coding: utf-8 -*-
"""
Calculations provided by aiida_flexpart.
Register calculations via the "aiida.calculations" entry point in setup.json.
"""
from pathlib import Path
from aiida import orm, common, engine
import yaml
import importlib

with importlib.resources.path('aiida_flexpart.templates', "params.yaml") as p:
    with open(p, 'r') as fp:
        params_dict = yaml.safe_load(fp)

cosmo_models = ['cosmo7', 'cosmo1', 'kenda1']
ECMWF_models = ['IFS_GL_05', 'IFS_GL_1', 'IFS_EU_02', 'IFS_EU_01']

class CollectSensitivitiesCalculation(engine.CalcJob):
    """AiiDA calculation plugin for the collection of sensitivities.
       The main input are the stashed Netcdf files for the previous FLEXPART
       calculations. """
    @classmethod
    def define(cls, spec):
        super().define(spec)

        #INPUTS metadata
        spec.inputs['metadata']['options']['resources'].default = {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        }
        spec.input('metadata.options.custom_scheduler_commands', valid_type=str, default='')
        spec.input('metadata.options.withmpi', valid_type=bool, default=False)
        spec.input('metadata.options.output_filename', valid_type=str, default='aiida.out', required=True)
        #spec.input('metadata.options.parser_name', valid_type=str, default='collect.sensitivities')

        #Inputs
        spec.input_namespace('remote', valid_type=orm.RemoteStashFolderData, required=True)

        #EXTRA INPUTS
        spec.input('model',valid_type = str,non_db=True, required = True)
        spec.input('outgrid',valid_type = str,non_db=True, required = True)
        spec.input('outgrid_n',valid_type = bool,non_db=True, required = True)

        #exit codes
        spec.outputs.dynamic = True
        spec.exit_code(300, 'ERROR_MISSING_OUTPUT_FILES', message='Calculation did not produce all expected output files.')

    def prepare_for_submission(self, folder):

        codeinfo = common.CodeInfo()
        codeinfo.cmdline_params = ['-p', 'params.yaml']
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.metadata.options.output_filename
        codeinfo.withmpi = self.inputs.metadata.options.withmpi

        with folder.open('params.yaml', 'w') as f:
            rel,path,days =[],[],[]
            for i,j in self.inputs.remote.items():
                rel.append(i[10:])
                path.append(j.target_basepath)
                days.append(i[:10].replace('_', '-'))
                
            str_ = 'IFS'
            if self.inputs.model in cosmo_models:
                str_='COSMO'
            
            params_dict.update({'rel.com':list(set(rel)),
                                'path':path,
                                'days':days,
                                'bs.path':path,
                                'domain.str':self.inputs.outgrid,
                                'nest':self.inputs.outgrid_n,
                                'met_model':self.inputs.model,
                                'model_version':'FLEXPART '+ str_
                                    })
            _ = yaml.dump(params_dict, f)

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = common.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.retrieve_list = ['aiida.out','*.nc']

        return calcinfo



# -*- coding: utf-8 -*-
"""Flexpart multi-dates WorkChain."""
from aiida import engine, plugins, orm
from aiida_shell import launch_shell_job
from aiida_flexpart.utils import get_simulation_period

#plugins
FlexpartCosmoCalculation = plugins.CalculationFactory('flexpart.cosmo')
FlexpartIfsCalculation = plugins.CalculationFactory('flexpart.ifs')
FlexpartPostCalculation = plugins.CalculationFactory('flexpart.post')

#possible models
cosmo_models = ['cosmo7', 'cosmo1', 'kenda1']
ECMWF_models = ['IFS_GL_05', 'IFS_GL_1', 'IFS_EU_02', 'IFS_EU_01']


class FlexpartMultipleDatesWorkflow(engine.WorkChain):
    """Flexpart multi-dates workflow"""
    @classmethod
    def define(cls, spec):
        """Specify inputs and outputs."""
        super().define(spec)

        #codes
        spec.input('fcosmo_code', valid_type=orm.AbstractCode)
        spec.input('check_meteo_cosmo_code', valid_type=orm.AbstractCode)
        spec.input('fifs_code', valid_type=orm.AbstractCode)
        spec.input('check_meteo_ifs_code', valid_type=orm.AbstractCode)
        spec.input('post_processing_code', valid_type=orm.AbstractCode)

        #extras
        spec.input('name', valid_type=str, non_db=True, required=False)

        # Basic Inputs
        spec.input('simulation_dates',
                   valid_type=orm.List,
                   help='A list of the starting dates of the simulations')
        spec.input('model', valid_type=orm.List, required=True)
        spec.input('model_offline', valid_type=orm.List, required=True)
        spec.input('offline_integration_time', valid_type=orm.Int)
        spec.input('integration_time',
                   valid_type=orm.Int,
                   help='Integration time in hours')
        spec.input(
            'parent_calc_folder',
            valid_type=orm.RemoteData,
            required=False,
            help=
            'Working directory of a previously ran calculation to restart from.'
        )

        #model settings
        spec.input('input_phy', valid_type=orm.Dict)
        spec.input('command', valid_type=orm.Dict)
        spec.input('release_settings', valid_type=orm.Dict)
        spec.input('locations',
                   valid_type=orm.Dict,
                   help='Dictionary of locations properties.')

        #meteo related inputs
        spec.input('meteo_inputs',
                   valid_type=orm.Dict,
                   required=False,
                   help='Meteo models input params.')
        spec.input('meteo_inputs_offline',
                   valid_type=orm.Dict,
                   required=False,
                   help='Meteo models input params.')
        spec.input(
            'meteo_path',
            valid_type=orm.List,
            required=False,
            help='Path to the folder containing the meteorological input data.'
        )
        spec.input(
            'meteo_path_offline',
            valid_type=orm.List,
            required=False,
            help='Path to the folder containing the meteorological input data.'
        )
        spec.input('gribdir', valid_type=orm.Str, required=True)

        #others
        spec.input('outgrid', valid_type=orm.Dict)
        spec.input('outgrid_nest', valid_type=orm.Dict, required=False)
        spec.input('species', valid_type=orm.RemoteData, required=True)
        spec.input_namespace('land_use',
                             valid_type=orm.RemoteData,
                             required=False,
                             dynamic=True,
                             help='#TODO')
        spec.input_namespace('land_use_ifs',
                             valid_type=orm.RemoteData,
                             required=False,
                             dynamic=True)

        spec.expose_inputs(FlexpartCosmoCalculation,
                           include=['metadata.options'],
                           namespace='flexpartcosmo')
        spec.expose_inputs(FlexpartIfsCalculation,
                           include=['metadata.options'],
                           namespace='flexpartifs')
        spec.expose_inputs(FlexpartPostCalculation,
                           include=['metadata.options'],
                           namespace='flexpartpost')

        # Outputs
        #spec.output('output_file', valid_type=orm.SinglefileData)
        spec.outputs.dynamic = True
        # What the workflow will do, step-by-step
        spec.outline(
            cls.setup,
            engine.while_(cls.condition)(
                engine.if_(cls.run_cosmo)(
                    engine.if_(cls.prepare_meteo_folder_cosmo)(
                        cls.run_cosmo_simulation)),
                engine.if_(cls.run_ifs)(
                    engine.if_(cls.prepare_meteo_folder_ifs)(
                        cls.run_ifs_simulation)),
                cls.post_processing,
            ),
            cls.results,
        )

    def condition(self):
        """multi dates loop"""
        return self.ctx.index < len(self.ctx.simulation_dates)

    def run_cosmo(self):
        """run cosmo simulation"""
        if all(mod in cosmo_models
               for mod in self.inputs.model) and self.inputs.model:
            return True
        return False

    def run_ifs(self):
        """run ifs simulation"""
        if (all(mod in ECMWF_models for mod in self.inputs.model)
                or all(mod in ECMWF_models
                       for mod in self.inputs.model_offline)
                and self.inputs.model and self.inputs.model_offline):
            return True
        return False

    def setup(self):
        """Prepare a simulation."""

        self.report('starting setup')

        self.ctx.index = 0
        self.ctx.simulation_dates = self.inputs.simulation_dates
        self.ctx.integration_time = self.inputs.integration_time
        self.ctx.offline_integration_time = self.inputs.offline_integration_time

        #model settings
        self.ctx.release_settings = self.inputs.release_settings
        self.ctx.command = self.inputs.command
        self.ctx.input_phy = self.inputs.input_phy
        self.ctx.locations = self.inputs.locations

        #others
        self.ctx.outgrid = self.inputs.outgrid
        self.ctx.species = self.inputs.species
        self.ctx.land_use = self.inputs.land_use
        if 'name' in self.inputs:
            out_n = 'None'
            if 'outgrid_nest' in self.inputs:
                out_n = self.inputs.outgrid_nest.get_dict()
            self.node.base.extras.set(
                self.inputs.name, {
                    'command': self.ctx.command.get_dict(),
                    'input_phy': self.ctx.input_phy.get_dict(),
                    'release': self.ctx.release_settings.get_dict(),
                    'locations': self.ctx.locations.get_dict(),
                    'integration_time': self.ctx.integration_time.value,
                    'offline_integration_time':
                    self.ctx.offline_integration_time.value,
                    'model': self.inputs.model,
                    'model_offline': self.inputs.model_offline,
                    'outgrid': self.inputs.outgrid.get_dict(),
                    'outgrid_nest': out_n
                })

    def prepare_meteo_folder_ifs(self):
        """prepare meteo folder"""
        age_class_ = self.inputs.integration_time.value * 3600
        if self.ctx.offline_integration_time > 0:
            age_class_ = self.inputs.offline_integration_time.value * 3600
        e_date, s_date = get_simulation_period(
            self.ctx.simulation_dates[self.ctx.index], age_class_,
            self.ctx.command.get_dict()['release_duration'],
            self.ctx.command.get_dict()['simulation_direction'])

        self.report(f'preparing meteo from {s_date} to {e_date}')

        if all(mod in ECMWF_models
               for mod in self.inputs.model) and self.inputs.model:
            model_list = self.inputs.model
        else:
            model_list = self.inputs.model_offline

        node_list = []
        for mod in model_list:
            self.report(f'transfering {mod} meteo')
            _, node = launch_shell_job(
                self.inputs.check_meteo_ifs_code,
                arguments=' -s {sdate} -e {edate} -g {gribdir} -m {model} -a',
                nodes={
                    'sdate': orm.Str(s_date),
                    'edate': orm.Str(e_date),
                    'gribdir': self.inputs.gribdir,
                    'model': orm.Str(mod)
                })
            node_list.append(node)

        if all(node.is_finished_ok for node in node_list):
            self.report('ALL meteo OK')
            return True

        self.report('FAILED to transfer meteo')
        self.ctx.index += 1
        return False

    def prepare_meteo_folder_cosmo(self):
        """prepare meteo folder"""
        e_date, s_date = get_simulation_period(
            self.ctx.simulation_dates[self.ctx.index],
            self.inputs.integration_time.value * 3600,
            self.ctx.command.get_dict()['release_duration'],
            self.ctx.command.get_dict()['simulation_direction'])

        self.report(f'preparing meteo from {s_date} to {e_date}')

        node_list = []
        for mod in self.inputs.model:
            self.report(f'transfering {mod} meteo')
            _, node = launch_shell_job(
                self.inputs.check_meteo_cosmo_code,
                arguments=' -s {sdate} -e {edate} -g {gribdir} -m {model} -a',
                nodes={
                    'sdate': orm.Str(s_date),
                    'edate': orm.Str(e_date),
                    'gribdir': self.inputs.gribdir,
                    'model': orm.Str(mod)
                })
            node_list.append(node)

        if all(node.is_finished_ok for node in node_list):
            self.report('ALL meteo OK')
            return True

        self.report('FAILED to transfer meteo')
        self.ctx.index += 1
        return False

    def post_processing(self):
        """post processing"""
        self.report('starting post-processsing')
        builder = FlexpartPostCalculation.get_builder()
        builder.code = self.inputs.post_processing_code
        builder.input_dir = self.ctx.calculations[-1].outputs.remote_folder

        if self.ctx.offline_integration_time > 0:
            self.report(
                f'main: {self.ctx.calculations[-2].outputs.remote_folder}')
            self.report(
                f'offline: {self.ctx.calculations[-1].outputs.remote_folder}')
            builder.input_dir = self.ctx.calculations[-2].outputs.remote_folder
            builder.input_offline_dir = self.ctx.calculations[
                -1].outputs.remote_folder

        builder.metadata.options = self.inputs.flexpartpost.metadata.options

        running = self.submit(builder)
        self.to_context(calculations=engine.append_(running))

    def run_cosmo_simulation(self):
        """Run calculations for equation of state."""

        self.report(
            f'starting flexpart cosmo {self.ctx.simulation_dates[self.ctx.index]}'
        )

        builder = FlexpartCosmoCalculation.get_builder()
        builder.code = self.inputs.fcosmo_code

        #update command file
        new_dict = self.ctx.command.get_dict()
        new_dict['simulation_date'] = self.ctx.simulation_dates[self.ctx.index]
        new_dict['age_class'] = self.inputs.integration_time * 3600
        new_dict.update(self.inputs.meteo_inputs)

        #model settings
        builder.model_settings = {
            'release_settings': self.ctx.release_settings,
            'locations': self.ctx.locations,
            'command': orm.Dict(dict=new_dict),
            'input_phy': self.ctx.input_phy,
        }

        builder.outgrid = orm.Dict(
            list(self.ctx.outgrid.get_dict().values())[0])
        if 'outgrid_nest' in self.inputs:
            builder.outgrid_nest = orm.Dict(
                list(self.inputs.outgrid_nest.get_dict().values())[0])
        builder.species = self.ctx.species
        builder.land_use = self.ctx.land_use
        builder.meteo_path = self.inputs.meteo_path

        # Walltime, memory, and resources.
        builder.metadata.description = 'Test workflow to submit a flexpart calculation'
        builder.metadata.options = self.inputs.flexpartcosmo.metadata.options

        # Ask the workflow to continue when the results are ready and store them in the context
        running = self.submit(builder)
        self.to_context(calculations=engine.append_(running))
        if self.ctx.offline_integration_time == 0:
            self.ctx.index += 1

    def run_ifs_simulation(self):
        """Run calculations for equation of state."""
        # Set up calculation.
        self.report(
            f'running flexpart ifs for {self.ctx.simulation_dates[self.ctx.index]}'
        )
        builder = FlexpartIfsCalculation.get_builder()
        builder.code = self.inputs.fifs_code

        #changes in the command file
        new_dict = self.ctx.command.get_dict()
        new_dict['simulation_date'] = self.ctx.simulation_dates[self.ctx.index]

        if self.ctx.offline_integration_time > 0:
            new_dict['age_class'] = self.ctx.offline_integration_time * 3600
            new_dict['dumped_particle_data'] = True

            self.ctx.parent_calc_folder = self.ctx.calculations[
                -1].outputs.remote_folder
            builder.parent_calc_folder = self.ctx.parent_calc_folder
            self.report(f'starting from: {self.ctx.parent_calc_folder}')

            builder.meteo_path = self.inputs.meteo_path_offline
            new_dict.update(self.inputs.meteo_inputs_offline)

        else:
            new_dict['age_class'] = self.inputs.integration_time * 3600
            builder.meteo_path = self.inputs.meteo_path
            new_dict.update(self.inputs.meteo_inputs)

        #model settings
        builder.model_settings = {
            'release_settings': self.ctx.release_settings,
            'locations': self.ctx.locations,
            'command': orm.Dict(dict=new_dict),
        }

        builder.outgrid = orm.Dict(
            list(self.ctx.outgrid.get_dict().values())[0])
        if 'outgrid_nest' in self.inputs:
            builder.outgrid_nest = orm.Dict(
                list(self.inputs.outgrid_nest.get_dict().values())[0])
        builder.species = self.ctx.species
        builder.land_use = self.inputs.land_use_ifs

        # Walltime, memory, and resources.
        builder.metadata.description = 'Test workflow to submit a flexpart calculation'
        builder.metadata.options = self.inputs.flexpartifs.metadata.options

        # Ask the workflow to continue when the results are ready and store them in the context
        running = self.submit(builder)
        self.to_context(calculations=engine.append_(running))

        self.ctx.index += 1

    def results(self):
        """Process results."""
        for indx, calculation in enumerate(self.ctx.calculations):
            self.out(f'calculation_{indx}_output_file',
                     calculation.outputs.output_file)



# -*- coding: utf-8 -*-
"""
Calculations provided by aiida_flexpart.

Register calculations via the "aiida.calculations" entry point in setup.json.
"""
import os
import importlib
import datetime
import pathlib
import jinja2

from aiida import common, orm, engine
from ..utils import fill_in_template_file


class FlexpartIfsCalculation(engine.CalcJob):
    """AiiDA calculation plugin wrapping the FLEXPART IFS executable."""
    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        # yapf: disable
        super().define(spec)

        # set default values for AiiDA options
        spec.inputs['metadata']['options']['resources'].default = {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        }

        spec.input('metadata.options.max_wallclock_seconds', valid_type=int, default=1800)
        spec.input('metadata.options.custom_scheduler_commands', valid_type=str, default='')
        spec.input('metadata.options.withmpi', valid_type=bool, default=False)
        spec.input('metadata.options.parser_name', valid_type=str, default='flexpart.ifs')

        spec.input(
            'parent_calc_folder',
            valid_type=orm.RemoteData,
            required=False,
            help='Working directory of a previously ran calculation to restart from.'
        )

        # Model settings
        spec.input_namespace('model_settings')
        spec.input('model_settings.release_settings', valid_type=orm.Dict, required=True)
        spec.input('model_settings.locations', valid_type=orm.Dict, required=True)
        spec.input('model_settings.command', valid_type=orm.Dict, required=True)

        spec.input('outgrid', valid_type=orm.Dict, help='Input file for the Lagrangian particle dispersion model FLEXPART.')
        spec.input('outgrid_nest', valid_type=orm.Dict, required=False,
            help='Input file for the Lagrangian particle dispersion model FLEXPART. Nested output grid.'
            )
        spec.input('species', valid_type=orm.RemoteData, required=True)
        spec.input_namespace('land_use', valid_type=orm.RemoteData, required=False, dynamic=True, help='#TODO')

        spec.input('meteo_path', valid_type=orm.List,
        required=True, help='Path to the folder containing the meteorological input data.')
        spec.input('metadata.options.output_filename', valid_type=str, default='aiida.out', required=True)
        spec.outputs.dynamic = True

        #exit codes
        spec.exit_code(300, 'ERROR_MISSING_OUTPUT_FILES', message='Calculation did not produce all expected output files.')

    @classmethod
    def _deal_with_time(cls, command_dict):
        """Dealing with simulation times."""
        #initial values
        simulation_beginning_date = datetime.datetime.strptime(command_dict.pop('simulation_date'),'%Y-%m-%d %H:%M:%S')
        age_class_time = datetime.timedelta(seconds=command_dict.pop('age_class'))
        release_chunk = datetime.timedelta(seconds=command_dict.pop('release_chunk'))
        release_duration = datetime.timedelta(seconds=command_dict.pop('release_duration'))

        #releases start and end times
        release_beginning_date=simulation_beginning_date
        release_ending_date=release_beginning_date+release_duration

        if command_dict['simulation_direction']>0: #forward
            simulation_ending_date=release_ending_date+age_class_time
        else: #backward
            simulation_ending_date=release_ending_date
            simulation_beginning_date-=age_class_time

        command_dict['simulation_beginning_date'] = [
            f'{simulation_beginning_date:%Y%m%d}',
            f'{simulation_beginning_date:%H%M%S}'
            ]
        command_dict['simulation_ending_date'] = [
            f'{simulation_ending_date:%Y%m%d}',
            f'{simulation_ending_date:%H%M%S}'
            ]
        return {
            'beginning_date': release_beginning_date,
            'ending_date': release_ending_date,
            'chunk': release_chunk
            } , age_class_time

    def prepare_for_submission(self, folder):

        meteo_string_list = ['./','./']
        for path in self.inputs.meteo_path:
            meteo_string_list.append(f'{path}{os.sep}')
            meteo_string_list.append(f'{path}/AVAILABLE')

        codeinfo = common.CodeInfo()
        codeinfo.cmdline_params = meteo_string_list
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.metadata.options.output_filename
        codeinfo.withmpi = self.inputs.metadata.options.withmpi

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = common.CalcInfo()
        calcinfo.codes_info = [codeinfo]


        command_dict = self.inputs.model_settings.command.get_dict()

        # Deal with simulation times.
        release, age_class_time = self._deal_with_time(command_dict)

        # Fill in the releases file.
        with folder.open('RELEASES', 'w') as infile:
            time_chunks = []
            current_time = release['beginning_date'] + release['chunk']
            while current_time <= release['ending_date']:
                time_chunks.append({
                    'begin': [f'{current_time-release["chunk"]:%Y%m%d}', f'{current_time-release["chunk"]:%H%M%S}'],
                    'end': [f'{current_time:%Y%m%d}', f'{current_time:%H%M%S}'],
                })
                current_time += release['chunk']

            template = jinja2.Template(importlib.resources.read_text('aiida_flexpart.templates', 'RELEASES.j2'))
            infile.write(template.render(
                time_chunks=time_chunks,
                locations=self.inputs.model_settings.locations.get_dict(),
                release_settings=self.inputs.model_settings.release_settings.get_dict()
                )
            )

        # Fill in the AGECLASSES file.
        fill_in_template_file(folder, 'AGECLASSES', int(age_class_time.total_seconds()))

        # Fill in the OUTGRID_NEST file if the corresponding dictionary is present.
        if 'outgrid_nest' in self.inputs:
            command_dict['nested_output'] = True
            fill_in_template_file(folder, 'OUTGRID_NEST_ifs', self.inputs.outgrid_nest.get_dict())
        else:
            command_dict['nested_output'] = False

        # Fill in the COMMAND file.
        fill_in_template_file(folder, 'COMMAND_ifs', command_dict)

        # Fill in the OUTGRID file.
        fill_in_template_file(folder, 'OUTGRID_ifs', self.inputs.outgrid.get_dict())


        calcinfo.remote_symlink_list = []
        calcinfo.remote_symlink_list.append((
            self.inputs.species.computer.uuid,
            self.inputs.species.get_remote_path(),
            'SPECIES'
            ))

        if 'parent_calc_folder' in self.inputs:
            computer_uuid = self.inputs.parent_calc_folder.computer.uuid
            remote_path = self.inputs.parent_calc_folder.get_remote_path()
            calcinfo.remote_symlink_list.append((
                         computer_uuid,
                         remote_path+'/header',
                         'header_previous'))
            calcinfo.remote_symlink_list.append((
                         computer_uuid,
                         remote_path+'/partposit_inst',
                         'partposit_previous'))


        # Dealing with land_use input namespace.
        for _, value in self.inputs.land_use.items():
            file_path = value.get_remote_path()
            calcinfo.remote_symlink_list.append((value.computer.uuid, file_path, pathlib.Path(file_path).name))

        calcinfo.retrieve_list = ['grid_time_*.nc', 'aiida.out']

        return calcinfo



# -*- coding: utf-8 -*-
"""
Calculations provided by aiida_flexpart.
Register calculations via the "aiida.calculations" entry point in setup.json.
"""
from aiida import orm, common, engine


class PostProcessingCalculation(engine.CalcJob):
    """AiiDA calculation plugin for post processing."""
    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        # yapf: disable
        super().define(spec)

        # set default values for AiiDA options
        spec.inputs['metadata']['options']['resources'].default = {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        }

        #INPUTS
        spec.input('metadata.options.custom_scheduler_commands', valid_type=str, default='')
        spec.input('metadata.options.withmpi', valid_type=bool, default=False)
        spec.input('metadata.options.parser_name', valid_type=str, default='flexpart.post')
        spec.input('input_dir', valid_type = orm.RemoteData, required=True,
                   help = 'main FLEXPART output dir')
        spec.input('input_offline_dir', valid_type = orm.RemoteData, required=False,
                   help = 'offline-nested FLEXPART output dir')
        spec.input('metadata.options.output_filename', valid_type=str, default='aiida.out', required=True)
        #exit codes
        spec.outputs.dynamic = True
        spec.exit_code(300, 'ERROR_MISSING_OUTPUT_FILES', message='Calculation did not produce all expected output files.')

    def prepare_for_submission(self, folder):

        params  = ['-m',self.inputs.input_dir.get_remote_path(),
                   '-r','./','-p'
                  ]
        if 'input_offline_dir' in self.inputs:
            params += ['-n',self.inputs.input_offline_dir.get_remote_path()]

        codeinfo = common.CodeInfo()
        codeinfo.cmdline_params = params
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.metadata.options.output_filename
        codeinfo.withmpi = self.inputs.metadata.options.withmpi

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = common.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.retrieve_list = ['grid_time_*.nc', 'boundary_sensitivity_*.nc','*.png', 'aiida.out']

        return calcinfo



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Run a multi dates workflow."""

import pathlib
import datetime
import click
import yaml
from aiida import orm, plugins, engine, cmdline
from aiida.common.datastructures import StashMode


def read_yaml_data(data_filename: str, names=None) -> dict:
    """Read in a YAML data file as a dictionary"""
    data_path = pathlib.Path(data_filename)
    with data_path.open('r', encoding='utf-8') as fp:
        yaml_data = yaml.safe_load(fp)

    return {key: value
            for key, value in yaml_data.items()
            if key in names} if names else yaml_data


def simulation_dates_parser(date_list: list) -> list:
    """
    Parse a range of dates and returns a list of date strings.

    Examples:
        2021-01-02--2021-01-10 -> [2021-01-02 00:00:00, 2021-01-02 00:00:00, ..., 2021-01-10 00:00:00]
        2021-01-02, 2021-01-10 -> [2021-01-02 00:00:00, 2021-01-10 00:00:00]
        2021-01-02 -> [2021-01-02 00:00:00,]
    """
    dates = []
    for date_string in date_list:
        if ',' in date_string:
            dates += [
                date.strip() + ' 00:00:00' for date in date_string.split(',')
            ]
        elif '--' in date_string:
            date_start, date_end = list(
                map(lambda date: datetime.datetime.strptime(date, '%Y-%m-%d'),
                    date_string.split('--')))
            dates += [
                date.strftime('%Y-%m-%d 00:00:00') for date in [
                    date_start + datetime.timedelta(days=x)
                    for x in range(0, (date_end - date_start).days + 1)
                ]
            ]
        else:
            dates += [date_string.strip() + ' 00:00:00']

    return orm.List(list=dates)


def test_run(flexpart_code):
    """Run workflow."""
    user_name='lfernand'
    simulation_dates = simulation_dates_parser(['2020-02-05'])

    # Links to the remote files/folders.
    glc = orm.RemoteData(remote_path=f'/users/{user_name}/resources/flexpart/GLC2000',
                         computer=flexpart_code.computer)
    species = orm.RemoteData(
        remote_path=f'/users/{user_name}/resources/flexpart/SPECIES',
        computer=flexpart_code.computer)
    surfdata = orm.RemoteData(
        remote_path=f'/users/{user_name}/resources/flexpart/surfdata.t',
        computer=flexpart_code.computer)
    surfdepo = orm.RemoteData(
        remote_path=f'/users/{user_name}/resources/flexpart/surfdepo.t',
        computer=flexpart_code.computer)
    meteo_path = orm.RemoteData(
        remote_path=f'/scratch/snx3000/{user_name}/FLEXPART_input/cosmo7/',
        computer=flexpart_code.computer)
    
    #builder starts
    workflow = plugins.WorkflowFactory('flexpart.multi_dates')
    builder = workflow.get_builder()
    builder.fcosmo_code = flexpart_code
    builder.check_meteo_cosmo_code =  orm.load_code('check-cosmo-data@daint-direct-106')
    builder.simulation_dates = simulation_dates

    builder.model=orm.Str('cosmo7')
    builder.meteo_path = meteo_path
    builder.gribdir=orm.Str(f'/scratch/snx3000/{user_name}/FLEXPART_input/')
    builder.outgrid = orm.Dict(
        dict=read_yaml_data('inputs/outgrid.yaml', names=[
            'Europe',
        ])['Europe'])
    builder.outgrid_nest = orm.Dict(dict=read_yaml_data(
        'inputs/outgrid_nest.yaml', names=[
            'Europe',
        ])['Europe'])  # optional input
   
    #model settings
    builder.command = orm.Dict(dict=read_yaml_data('inputs/command.yaml'))
    builder.input_phy = orm.Dict(dict=read_yaml_data('inputs/input_phy.yaml'))
    builder.release_settings = orm.Dict(dict=read_yaml_data('inputs/release.yaml'))
    builder.locations = orm.Dict(
        dict=read_yaml_data('inputs/locations.yaml', names=[
            'TEST_32',
        ]))
    
    builder.meteo_inputs = orm.Dict(
        dict=read_yaml_data('inputs/meteo_inputs.yaml', names=[
            'cosmo7',
        ])['cosmo7'])
    builder.integration_time = orm.Int(24)
    builder.species = species
    builder.land_use = {
        'glc': glc,
        'surfdata': surfdata,
        'surfdepo': surfdepo,
    }

    builder.flexpart.metadata.options.stash = {
        'source_list': ['aiida.out','partposit_inst', 'header', 'grid_time_*.nc'],
        'target_base': f'/store/empa/em05/{user_name}/aiida_stash',
        'stash_mode': StashMode.COPY.value,
    }
    engine.run(builder)


@click.command()
@cmdline.utils.decorators.with_dbenv()
@cmdline.params.options.CODE()
def cli(code):
    """Run example.

    Example usage: $ ./example_01.py --code diff@localhost

    Alternative (creates diff@localhost-test code): $ ./example_01.py

    Help: $ ./example_01.py --help
    """
    test_run(code)


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter


# -*- coding: utf-8 -*-
"""
Calculations provided by aiida_flexpart.

Register calculations via the "aiida.calculations" entry point in setup.json.
"""
import os
import importlib
import datetime
import pathlib
import jinja2

from aiida import orm
from aiida.common import datastructures
from aiida.engine import CalcJob
from aiida_flexpart.utils import convert_input_to_namelist_entry

from ..utils import fill_in_template_file


class FlexpartCosmoCalculation(CalcJob):
    """AiiDA calculation plugin wrapping the FLEXPART executable."""
    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        # yapf: disable
        super().define(spec)

        # set default values for AiiDA options
        spec.inputs['metadata']['options']['resources'].default = {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        }

        spec.input('metadata.options.max_wallclock_seconds', valid_type=int, default=1800)
        spec.input('metadata.options.custom_scheduler_commands', valid_type=str, default='')
        spec.input('metadata.options.withmpi', valid_type=bool, default=False)
        spec.input('metadata.options.parser_name', valid_type=str, default='flexpart.cosmo')

        # new ports
        # Model settings namespace.
        spec.input_namespace('model_settings')
        spec.input('model_settings.release_settings', valid_type=orm.Dict, required=True)
        spec.input('model_settings.locations', valid_type=orm.Dict, required=True)
        spec.input('model_settings.command', valid_type=orm.Dict, required=True)
        spec.input('model_settings.input_phy',  valid_type=orm.Dict, required=True)

        spec.input('outgrid', valid_type=orm.Dict, help='Input file for the Lagrangian particle dispersion model FLEXPART.')
        spec.input(
            'outgrid_nest',
            valid_type=orm.Dict,
            required=False,
            help='Input file for the Lagrangian particle dispersion model FLEXPART. Nested output grid.'
            )
        spec.input('species', valid_type=orm.RemoteData, required=True)
        spec.input('meteo_path', valid_type=orm.List,
        required=True, help='Path to the folder containing the meteorological input data.')
        spec.input('metadata.options.output_filename', valid_type=str, default='aiida.out', required=True)
        spec.input_namespace('land_use', valid_type=orm.RemoteData, required=False, dynamic=True, help='#TODO')

        spec.outputs.dynamic = True
        spec.exit_code(300, 'ERROR_MISSING_OUTPUT_FILES', message='Calculation did not produce all expected output files.')

    @classmethod
    def _deal_with_time(cls, command_dict):
        """Dealing with simulation times."""

        #initial values
        simulation_beginning_date = datetime.datetime.strptime(command_dict.pop('simulation_date'),'%Y-%m-%d %H:%M:%S')
        age_class_time = datetime.timedelta(seconds=command_dict.pop('age_class'))
        release_chunk = datetime.timedelta(seconds=command_dict.pop('release_chunk'))
        release_duration = datetime.timedelta(seconds=command_dict.pop('release_duration'))

        #releases start and end times
        release_beginning_date=simulation_beginning_date
        release_ending_date=release_beginning_date+release_duration

        if command_dict['simulation_direction']>0: #forward
            simulation_ending_date=release_ending_date+age_class_time
        else: #backward
            simulation_ending_date=release_ending_date
            simulation_beginning_date-=age_class_time

        command_dict['simulation_beginning_date'] = [
            f'{simulation_beginning_date:%Y%m%d}',
            f'{simulation_beginning_date:%H%M%S}'
            ]
        command_dict['simulation_ending_date'] = [
            f'{simulation_ending_date:%Y%m%d}',
            f'{simulation_ending_date:%H%M%S}'
            ]
        return {
            'beginning_date': release_beginning_date,
            'ending_date': release_ending_date,
            'chunk': release_chunk
            } , age_class_time

    def prepare_for_submission(self, folder):  # pylint: disable=too-many-locals
        """
        Create input files.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily place all files
            needed by the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """
        meteo_string_list = ['./','./']
        for path in self.inputs.meteo_path:
            meteo_string_list.append(f'{path}{os.sep}')
            meteo_string_list.append(f'{path}/AVAILABLE')

        codeinfo = datastructures.CodeInfo()
        codeinfo.cmdline_params = meteo_string_list
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.metadata.options.output_filename
        codeinfo.withmpi = self.inputs.metadata.options.withmpi

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]

        # Convert input_phy dictionary to the INPUT_PHY input file
        with folder.open('INPUT_PHY', 'w') as infile:
            infile.write('&parphy\n')
            for key, value in self.inputs.model_settings.input_phy.get_dict().items():
                infile.write(convert_input_to_namelist_entry(key, value))
            infile.write('/\n')

        command_dict = self.inputs.model_settings.command.get_dict()

        # Deal with simulation times.
        release, age_class_time = self._deal_with_time(command_dict)

        # Fill in the releases file.
        with folder.open('RELEASES', 'w') as infile:
            time_chunks = []
            current_time = release['beginning_date'] + release['chunk']
            while current_time <= release['ending_date']:
                time_chunks.append({
                    'begin': [f'{current_time-release["chunk"]:%Y%m%d}', f'{current_time-release["chunk"]:%H%M%S}'],
                    'end': [f'{current_time:%Y%m%d}', f'{current_time:%H%M%S}'],
                })
                current_time += release['chunk']

            template = jinja2.Template(importlib.resources.read_text('aiida_flexpart.templates', 'RELEASES.j2'))
            infile.write(template.render(
                time_chunks=time_chunks,
                locations=self.inputs.model_settings.locations.get_dict(),
                release_settings=self.inputs.model_settings.release_settings.get_dict()
                )
            )

        # Fill in the AGECLASSES file.
        fill_in_template_file(folder, 'AGECLASSES', int(age_class_time.total_seconds()))

        # Fill in the OUTGRID_NEST file if the corresponding dictionary is present.
        if 'outgrid_nest' in self.inputs:
            command_dict['nested_output'] = True
            fill_in_template_file(folder, 'OUTGRID_NEST', self.inputs.outgrid_nest.get_dict())
        else:
            command_dict['nested_output'] = False

        # Fill in the COMMAND file.
        fill_in_template_file(folder, 'COMMAND', command_dict)

        # Fill in the OUTGRID file.
        fill_in_template_file(folder, 'OUTGRID', self.inputs.outgrid.get_dict())

        calcinfo.remote_symlink_list = []
        calcinfo.remote_symlink_list.append((
            self.inputs.species.computer.uuid,
            self.inputs.species.get_remote_path(),
            'SPECIES'
            ))

        # Dealing with land_use input namespace.
        for _, value in self.inputs.land_use.items():
            file_path = value.get_remote_path()
            calcinfo.remote_symlink_list.append((value.computer.uuid, file_path, pathlib.Path(file_path).name))

        calcinfo.retrieve_list = ['grid_time_*.nc', 'aiida.out']

        return calcinfo



# -*- coding: utf-8 -*-
"""Check that version numbers match.

Check version number in setup.json and aiida_flexpart/__init__.py and make sure
they match.
"""
from __future__ import absolute_import
from __future__ import print_function
import os
import json
import sys

this_path = os.path.split(os.path.realpath(__file__))[0]

# Get content of setup.json
SETUP_FNAME = 'setup.json'
SETUP_PATH = os.path.join(this_path, os.pardir, SETUP_FNAME)
with open(SETUP_PATH) as f:
    setup_content = json.load(f)

# Get version from python package
sys.path.insert(0, os.path.join(this_path, os.pardir))
import aiida_flexpart  # pylint: disable=wrong-import-position
VERSION = aiida_flexpart.__version__

if VERSION != setup_content['version']:
    print('Version number mismatch detected:')
    print("Version number in '{}': {}".format(SETUP_FNAME,
                                              setup_content['version']))
    print("Version number in '{}/__init__.py': {}".format(
        'aiida_flexpart', VERSION))
    sys.exit(1)

# Overwrite version in setup.json
#setup_content['version'] = version
#with open(SETUP_PATH, 'w') as f:
#	json.dump(setup_content, f, indent=4, sort_keys=True)



# -*- coding: utf-8 -*-
""" Helper functions for automatically setting up computer & code.
Helper functions for setting up

 1. An AiiDA localhost computer
 2. A "diff" code on localhost

Note: Point 2 is made possible by the fact that the ``diff`` executable is
available in the PATH on almost any UNIX system.
"""
import tempfile
import shutil
from aiida.orm import Computer, Code
from aiida.common.exceptions import NotExistent

LOCALHOST_NAME = 'localhost-test'

executables = {
    'flexpart.cosmo': 'FLEXPART_master_daint_GNUoptompncdfout',
}


def get_path_to_executable(executable):
    """ Get path to local executable.
    :param executable: Name of executable in the $PATH variable
    :type executable: str
    :return: path to executable
    :rtype: str
    """
    path = shutil.which(executable)
    if path is None:
        raise ValueError(
            "'{}' executable not found in PATH.".format(executable))
    return path


def get_computer(name=LOCALHOST_NAME, workdir=None):
    """Get AiiDA computer.
    Loads computer 'name' from the database, if exists.
    Sets up local computer 'name', if it isn't found in the DB.

    :param name: Name of computer to load or set up.
    :param workdir: path to work directory
        Used only when creating a new computer.
    :return: The computer node
    :rtype: :py:class:`aiida.orm.Computer`
    """

    try:
        computer = Computer.objects.get(label=name)
    except NotExistent:
        if workdir is None:
            workdir = tempfile.mkdtemp()

        computer = Computer(
            label=name,
            description='localhost computer set up by aiida_diff tests',
            hostname=name,
            workdir=workdir,
            transport_type='local',
            scheduler_type='direct')
        computer.store()
        computer.set_minimum_job_poll_interval(0.)
        computer.configure()

    return computer


def get_code(entry_point, computer):
    """Get local code.
    Sets up code for given entry point on given computer.

    :param entry_point: Entry point of calculation plugin
    :param computer: (local) AiiDA computer
    :return: The code node
    :rtype: :py:class:`aiida.orm.Code`
    """

    try:
        executable = executables[entry_point]
    except KeyError as exc:
        raise KeyError(
            "Entry point '{}' not recognized. Allowed values: {}".format(
                entry_point, list(executables.keys()))) from exc

    codes = Code.objects.find(filters={'label': executable})  # pylint: disable=no-member
    if codes:
        return codes[0]

    path = get_path_to_executable(executable)
    code = Code(
        input_plugin_name=entry_point,
        remote_computer_exec=[computer, path],
    )
    code.label = executable
    return code.store()



# -*- coding: utf-8 -*-
from aiida import engine, orm
from aiida_shell import launch_shell_job
from aiida_flexpart.utils import get_simulation_period

#possible models
cosmo_models = ['cosmo7', 'cosmo1', 'kenda1']
ECMWF_models = ['IFS_GL_05', 'IFS_GL_1', 'IFS_EU_02', 'IFS_EU_01']


class TransferMeteoWorkflow(engine.WorkChain):
    """Multi-dates workflow for transfering the necessary
    Meteorological data for the subsequent Flexpart
    simulations."""
    @classmethod
    def define(cls, spec):
        """Specify inputs and outputs."""
        super().define(spec)

        spec.input('check_meteo_cosmo_code', valid_type=orm.AbstractCode)
        spec.input('check_meteo_ifs_code', valid_type=orm.AbstractCode)

        spec.input('simulation_dates', valid_type=orm.List, required=False)
        spec.input('gribdir', valid_type=orm.Str, required=True)

        spec.input('model', valid_type=orm.List, required=True)
        spec.input('model_offline', valid_type=orm.List, required=True)
        spec.input('offline_integration_time', valid_type=orm.Int)
        spec.input('integration_time', valid_type=orm.Int)
        spec.input('command', valid_type=orm.Dict)

        spec.outline(
            cls.setup,
            engine.while_(cls.date)(cls.prepare_meteo_folder,
                                    engine.if_(cls.offline)(
                                        cls.prepare_meteo_folder),
                                    cls.add_index))

    def setup(self):
        self.ctx.index = 0

    def date(self):
        self.ctx.offline = False
        return self.ctx.index < len(self.inputs.simulation_dates)

    def offline(self):
        if self.inputs.offline_integration_time > 0:
            self.ctx.offline = True
            return True
        return False

    def add_index(self):
        self.ctx.index += 1

    def prepare_meteo_folder(self):
        model_list = self.inputs.model
        code_ = self.inputs.check_meteo_cosmo_code
        age_class_ = self.inputs.integration_time.value * 3600

        if all(mod in ECMWF_models
               for mod in self.inputs.model) and self.inputs.model:
            code_ = self.inputs.check_meteo_ifs_code

        if self.ctx.offline > 0:
            age_class_ = self.inputs.offline_integration_time.value * 3600
            model_list = self.inputs.model_offline
            code_ = self.inputs.check_meteo_ifs_code

        e_date, s_date = get_simulation_period(
            self.inputs.simulation_dates[self.ctx.index],
            age_class_,
            self.inputs.command.get_dict()['release_duration'],
            self.inputs.command.get_dict()['simulation_direction'],
        )
        self.report(f'preparing meteo from {s_date} to {e_date}')

        node_list = []
        for mod in model_list:
            self.report(f'transfering {mod} meteo')
            _, node = launch_shell_job(
                code_,
                arguments=' -s {sdate} -e {edate} -g {gribdir} -m {model} -a',
                nodes={
                    'sdate': orm.Str(s_date),
                    'edate': orm.Str(e_date),
                    'gribdir': self.inputs.gribdir,
                    'model': orm.Str(mod),
                },
            )
            node_list.append(node)

        if all(node.is_finished_ok for node in node_list):
            self.report('ALL meteo OK')



# -*- coding: utf-8 -*-
"""Utilties to convert between python and fortran data types and formats."""

import numbers
import datetime
import importlib
import numpy
import jinja2


def conv_to_fortran(val, quote_strings=True):
    """Convert a python value to a format suited for fortran input.
    :param val: the value to be read and converted to a Fortran-friendly string.
    """

    # Note that bool should come before integer, because a boolean matches also isinstance(..., int)
    if isinstance(val, (bool, numpy.bool_)):
        if val:
            val_str = '.true.'
        else:
            val_str = '.false.'
    elif isinstance(val, numbers.Integral):
        val_str = f'{val:d}'
    elif isinstance(val, numbers.Real):
        val_str = f'{val:18.10e}'.replace('e', 'd')
    elif isinstance(val, str):
        if quote_strings:
            val_str = f"'{val!s}'"
        else:
            val_str = f'{val!s}'
    else:
        raise ValueError(
            f"Invalid value '{val}' of type '{type(val)}' passed, accepts only bools, ints, floats and strings"
        )

    return val_str


def convert_input_to_namelist_entry(key, val, mapping=None):
    """Convert a key and a value, from an input parameters dictionary for a namelist calculation.
    Map it to the  appropriate string format for the namelist input file. For single values it will return a single
    string, but for values that are a dictionary, list or tuple, the returned string may be multiline.
    :param key: the namelist keyword name
    :param val: the namelist keyword value
        The value can be either a single value, list/tuple, a double nested list or a dictionary.
        Depending on the type of the value the resulting string differs vastly
        * single list:
            A list of keywords will be generated, where the index of the value in the list will be
            used as the index in the keyword and the value itself will be converted using conv_to_fortran.
            For example::
                'efield': [4, 5, 6]
            will result in::
                efield(1) = 4
                efield(1) = 5
                efield(1) = 6
        * double nested list:
            This format can be used for keywords that require one or more indices that do not necessarily
            follow a sequential number, but take specific values that need to be defined by the user.
            For example::
                'starting_ns_eigenvalue': [
                    [1, 1, 3, 3.5],
                    [2, 1, 1, 2.8]
                ]
            will be formatted as::
                starting_ns_eigenvalue(1,1,3) = 3.5
                starting_ns_eigenvalue(2,1,1) = 2.8
            Note that if the mapping argument is provided in the input, any value in sub lists that matches
            a key in the mapping dictionary (that is to say it is a string that matches one of the kinds), it
            will be replaced with the index of the corresponding atomic species. For example::
                hubbard_j: [
                    [2, 'Ni', 3.5],
                    [2, 'Fe', 7.4],
                ]
            would be formatted as::
                hubbard_j(2, 1) = 3.5
                hubbard_j(2, 3) = 7.4
            Assuming the mapping dictionary contained the kinds 'Ni' and 'Fe', with the indices 1 and 3, respectively
        * dictionary:
            The keys of this dictionary should correspond to keys in the mapping input argument and will be replaced
            with the corresponding value. This can be used for keywords that take a single index that needs to conform
            to the index of the atomic species to which the keyword value should apply. For example::
                hubbard_u: {
                    'Co': 3.5,
                    'O': 7.4,
                }
            will be formatted as::
                hubbard_u(1) = 3.5
                hubbard_u(3) = 7.4
            assuming that the kinds 'Co' and 'O' would have atomic species indices 1 and 3, respectively.
            This mapping from kind name to atomic species index should be defined by the `mapping` argument.
    :param mapping: optional parameter, that must be provided if val is a dictionary or a double nested list
        where the sub lists contain string values. The keys of the mapping dictionary should be the atomic species
        names that will be encountered in the value, and the corresponding value should be the index of that
        atomic species. Example::
            mapping = {
                'Fe': 1,
                'O': 2,
            }
        This will map every occurrence of 'Fe' and 'O' in the values to the corresponding integer.
    """
    # pylint: disable=too-many-branches,too-many-nested-blocks,no-else-return
    # I don't try to do iterator=iter(val) and catch TypeError because it would also match strings
    # I check first the dictionary, because it would also match hasattr(__iter__)
    if isinstance(val, dict):

        if mapping is None:
            raise ValueError(
                "If 'val' is a dictionary, you must provide also the 'mapping' parameter"
            )

        # At difference with the case of a list, at the beginning list_of_strings
        # is a list of 2-tuples where the first element is the idx, and the
        # second is the actual line. This is used at the end to resort everything.
        list_of_strings = []

        for elemk, itemval in val.items():
            try:
                idx = mapping[elemk]
            except KeyError as exception:
                raise ValueError(
                    f"Unable to find the key '{elemk}' in the mapping dictionary"
                ) from exception

            list_of_strings.append(
                (idx, f'  {key}({idx}) = {conv_to_fortran(itemval)}\n'))

        # I first have to resort, then to remove the index from the first column, finally to join the strings
        list_of_strings = list(zip(*sorted(list_of_strings)))[1]
        return ''.join(list_of_strings)

    # A list/tuple of values
    elif isinstance(val, (list, tuple)):

        list_of_strings = []

        for idx, itemval in enumerate(val):

            if isinstance(itemval, (list, tuple)):

                values = []

                for value in itemval[:-1]:

                    if not isinstance(value, (int, str)):
                        raise ValueError(
                            'values of double nested lists should be either integers or strings'
                        )

                    if isinstance(value, str):
                        if mapping is None:
                            raise ValueError(
                                'cannot map the string value because no mapping was defined'
                            )

                        if value not in mapping:
                            raise ValueError(
                                f'the nested list contained string {value} but this is not a key in the mapping'
                            )
                        values.append(str(mapping[value]))
                    else:
                        values.append(str(value))

                idx_string = ','.join(values)
                itemval = itemval.pop()
            else:
                idx_string = f'{idx + 1}'

            list_of_strings.append(
                f'  {key}({idx_string}) = {conv_to_fortran(itemval)}\n')

        return ''.join(list_of_strings)

    # Single value
    else:
        return f'  {key} = {conv_to_fortran(val)}\n'


def fill_in_template_file(folder, fname, data):
    """Create an input file based on the standard templates."""

    if 'ifs' in fname:
        fname_ = fname[:-4]
    else:
        fname_ = fname

    with folder.open(fname_, 'w') as infile:
        template = jinja2.Template(
            importlib.resources.read_text('aiida_flexpart.templates',
                                          fname + '.j2'))
        infile.write(template.render(data=data))


def reformat_locations(dict_, model):
    """reformat locations"""
    for key in dict_.keys():
        if 'longitude' in dict_[key]:
            dict_[key]['longitude_of_lower_left_corner'] = dict_[key][
                'longitude']
            dict_[key]['longitude_of_upper_right_corner'] = dict_[key][
                'longitude']
            dict_[key]['latitude_of_lower_left_corner'] = dict_[key][
                'latitude']
            dict_[key]['latitude_of_upper_right_corner'] = dict_[key][
                'latitude']

            if model in dict_[key]['level']:
                dict_[key]['lower_z_level'] = dict_[key]['level'][model]
                dict_[key]['upper_z_level'] = dict_[key]['level'][model]
            else:
                dict_[key]['lower_z_level'] = dict_[key]['level']['default']
                dict_[key]['upper_z_level'] = dict_[key]['level']['default']

            dict_[key]['level_type'] = dict_[key]['level_type'][model]

            dict_[key].pop('longitude')
            dict_[key].pop('latitude')
            dict_[key].pop('level')
    return dict_


def get_simulation_period(date, age_class_time, release_duration,
                          simulation_direction):
    """Dealing with simulation times."""
    #initial values
    simulation_beginning_date = datetime.datetime.strptime(
        date, '%Y-%m-%d %H:%M:%S')
    age_class_time = datetime.timedelta(seconds=age_class_time)
    release_duration = datetime.timedelta(seconds=release_duration + 3600)

    if simulation_direction > 0:  #forward
        simulation_ending_date = simulation_beginning_date + release_duration + age_class_time
    else:  #backward
        simulation_ending_date = release_duration + simulation_beginning_date
        simulation_beginning_date -= age_class_time

    return datetime.datetime.strftime(simulation_ending_date,
                                      '%Y%m%d%H'), datetime.datetime.strftime(
                                          simulation_beginning_date,
                                          '%Y%m%d%H')



# -*- coding: utf-8 -*-
from aiida import orm, common, engine
from aiida.plugins import DataFactory
import yaml

NetCDF = DataFactory('netcdf.data')


class Inversion(engine.CalcJob):
    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        # yapf: disable
        super().define(spec)

        #INPUTS metadata
        spec.inputs['metadata']['options']['resources'].default = {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        }
        spec.input('metadata.options.custom_scheduler_commands', valid_type=str, default='')
        spec.input('metadata.options.withmpi', valid_type=bool, default=False)
        spec.input('metadata.options.output_filename', valid_type=str, default='aiida.out', required=True)
        spec.input('metadata.options.parser_name', valid_type=str, default='inversion.calc')

        #Inputs
        spec.input_namespace('remotes', valid_type = NetCDF, required=True)
        spec.input_namespace('observations', valid_type = NetCDF, required=True)

        spec.input('inv_params',valid_type = orm.Dict, required = True,
                   help = 'File containing inversion settings, either as R source file or yaml')
        spec.input('start_date',valid_type = orm.Str, required = True,
                   help = 'Start date (yyyy-mm-dd)')
        spec.input('end_date',valid_type = orm.Str, required = True,
                   help = 'End date (yyyy-mm-dd)')
        spec.input('chunk',valid_type = orm.Str, required = True,
                   help = "Options are 'year' and 'month'. Default is 'year'")
        spec.input('chunk_w',valid_type = orm.Str, required = True,
                   help = """Width of the individual inversion chunk. These can be wider than
			                 the chunking itself to allow for running average fluxes.,
                             Possible values are 'year' and '3year' for 'chunk.by=year' and,
                             'month' and '3month' for 'chunk.by=month'. Default is 'year'
                         """)

        #exit codes
        spec.outputs.dynamic = True

    def prepare_for_submission(self, folder):

        codeinfo = common.CodeInfo()
        codeinfo.cmdline_params = [
            '-f',
            'inversion_settings.yaml',
            '-s',
            self.inputs.start_date.value,
            '-e',
            self.inputs.end_date.value,
            '-c',
            self.inputs.chunk.value,
            '-w',
            self.inputs.chunk_w.value,
        ]
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.metadata.options.output_filename
        codeinfo.withmpi = self.inputs.metadata.options.withmpi


        #create dict for yaml, add remotes by location
        remote_dict = {}
        for k,v in self.inputs.remotes.items():
             if k.split("_")[0] in remote_dict.keys():
                remote_dict[k.split("_")[0]].append(v.attributes["remote_path"])
             else:
                remote_dict[k.split("_")[0]] = [v.attributes["remote_path"]]

        params_dict = self.inputs.inv_params.get_dict()
        for k,v in remote_dict.items():
            params_dict['sites'][k].update({'ft.fls':v})

        #replace _ by . in dict
        params_dict = {
    key.replace("_", "."): value for key, value in params_dict.items()
}

        with folder.open('inversion_settings.yaml', 'w') as f:
                _ = yaml.dump(params_dict, f)
            
        calcinfo = common.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.retrieve_list = ['aiida.out']

        return calcinfo



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""ifs example run"""

import pathlib
import click
import yaml
from aiida import cmdline, engine, orm, plugins, common

INPUT_DIR = pathlib.Path(__file__).resolve().parent / 'input_files'


def read_yaml_data(data_filename: str, names=None) -> dict:
    """Read in a YAML data file as a dictionary"""
    data_path = pathlib.Path(data_filename)
    with data_path.open('r', encoding='utf-8') as fp:
        yaml_data = yaml.safe_load(fp)

    return {key: value
            for key, value in yaml_data.items()
            if key in names} if names else yaml_data


def test_run(flexpart_code):
    """Run a calculation on the localhost computer.

    Uses test helpers to create AiiDA Code on the fly.
    """
    user_name = 'lfernand'
    # Prepare input parameters

    command = orm.Dict(dict=read_yaml_data('inputs/command.yaml'))

    outgrid = orm.Dict(
        dict=read_yaml_data('inputs/outgrid.yaml', names=['Europe'])['Europe'])

    outgrid_nest = orm.Dict(dict=read_yaml_data('inputs/outgrid_nest.yaml',
                                                names=['Europe'])['Europe'])

    release_settings = orm.Dict(dict=read_yaml_data('inputs/release.yaml'))

    locations = orm.Dict(dict=read_yaml_data('inputs/locations.yaml',
                                             names=['TEST_32', 'TEST_200']))

    # Links to the remote files/folders.
    glc = orm.RemoteData(
        remote_path=f'/users/{user_name}/resources/flexpart/IGBP_int1.dat',
        computer=flexpart_code.computer)
    species = orm.RemoteData(
        remote_path=f'/users/{user_name}/resources/flexpart/SPECIES',
        computer=flexpart_code.computer)
    surfdata = orm.RemoteData(
        remote_path=f'/users/{user_name}/resources/flexpart/surfdata.t',
        computer=flexpart_code.computer)
    surfdepo = orm.RemoteData(
        remote_path=f'/users/{user_name}/resources/flexpart/surfdepo.t',
        computer=flexpart_code.computer)
    meteo_path = orm.RemoteData(
        remote_path=f'/scratch/snx3000/{user_name}/FLEXPART_input/IFS_GL_05',
        computer=flexpart_code.computer)

    # Set up calculation.
    calc = plugins.CalculationFactory('flexpart.ifs')
    builder = calc.get_builder()
    builder.code = flexpart_code
    builder.model_settings = {
        'release_settings': release_settings,
        'locations': locations,
        'command': command,
    }
    builder.outgrid = outgrid
    builder.outgrid_nest = outgrid_nest
    builder.species = species
    builder.meteo_path = meteo_path
    builder.land_use = {
        'glc': glc,
        'surfdata': surfdata,
        'surfdepo': surfdepo,
    }

    #uncomment to use previous cosmo calculation
    #builder.parent_calc_folder = previous_cosmo_calc

    builder.metadata.description = 'Test job submission with the aiida_flexpart plugin'
    builder.metadata.options.stash = {
        'source_list': ['aiida.out', 'grid_time_*.nc'],
        'target_base': f'/store/empa/em05/{user_name}/aiida_stash',
        'stash_mode': common.StashMode.COPY.value,
    }
    #builder.metadata.options.max_wallclock_seconds = 2000

    # builder.metadata.dry_run = True
    # builder.metadata.store_provenance = False
    engine.run(builder)
    # result = engine.submit(builder) # submit to aiida daemon


@click.command()
@cmdline.utils.decorators.with_dbenv()
@cmdline.params.options.CODE()
def cli(code):
    """Run example.

    Example usage: $ ./example_01.py --code diff@localhost

    Alternative (creates diff@localhost-test code): $ ./example_01.py

    Help: $ ./example_01.py --help
    """
    test_run(code)


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-
"""pytest fixtures for simplified testing."""
from __future__ import absolute_import
import pytest
pytest_plugins = ['aiida.manage.tests.pytest_fixtures']


@pytest.fixture(scope='function', autouse=True)
def clear_database_auto(clear_database):  # pylint: disable=unused-argument
    """Automatically clear database in between tests."""


@pytest.fixture(scope='function')
def flexpart_code(aiida_local_code_factory):
    """Get a flexpart code.
    """
    return aiida_local_code_factory(executable='diff', entry_point='flexpart')



# -*- coding: utf-8 -*-
"""
Parsers provided by aiida_flexpart.

Register parsers via the "aiida.parsers" entry point in setup.json.
"""
from aiida.engine import ExitCode
from aiida.parsers.parser import Parser
from aiida.plugins import CalculationFactory
from aiida.common import exceptions
from aiida.orm import SinglefileData

InversionCalculation = CalculationFactory('inversion.calc')


class InvesrionParser(Parser):
    """
    Parser class for parsing output of calculation.
    """
    def __init__(self, node):
        """
        Initialize Parser instance

        Checks that the ProcessNode being passed was produced by a FlexpartCalculation.

        :param node: ProcessNode of calculation
        :param type node: :class:`aiida.orm.ProcessNode`
        """
        super().__init__(node)
        if not issubclass(node.process_class, InversionCalculation):
            raise exceptions.ParsingError('Can only parse FlexpartCalculation')

    def parse(self, **kwargs):
        """
        Parse outputs, store results in database.

        :returns: an exit code, if parsing fails (or nothing if parsing succeeds)
        """
        output_filename = self.node.get_option('output_filename')

        # Check that folder content is as expected
        files_retrieved = self.retrieved.list_object_names()
        files_expected = [output_filename]
        # Note: set(A) <= set(B) checks whether A is a subset of B
        if not set(files_expected) <= set(files_retrieved):
            self.logger.error("Found files '{}', expected to find '{}'".format(
                files_retrieved, files_expected))
            return self.exit_codes.ERROR_MISSING_OUTPUT_FILES

        # add output file
        self.logger.info(f"Parsing '{output_filename}'")
        with self.retrieved.open(output_filename, 'r') as handle:
            content = handle.read()
            output_node = SinglefileData(file=handle)
            if 'CONGRATULATIONS' not in content:
                self.out('output_file', output_node)
                return ExitCode(1)

        self.out('output_file', output_node)
        return ExitCode(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Run a multi dates workflow."""

import pathlib
import datetime
import click
import yaml
from aiida import orm, plugins, engine, cmdline, common
from aiida_flexpart.utils import reformat_locations


def read_yaml_data(data_filename: str, names=None) -> dict:
    """Read in a YAML data file as a dictionary"""
    data_path = pathlib.Path(data_filename)
    with data_path.open('r', encoding='utf-8') as fp:
        yaml_data = yaml.safe_load(fp)

    return {key: value
            for key, value in yaml_data.items()
            if key in names} if names else yaml_data


def make_locations_list(list_locations):
    """makes location list"""
    list_locations_ = read_yaml_data('inputs/location_groups.yaml',
                                     names=list_locations)
    list_ = []
    if list_locations_:
        for _, j in list_locations_.items():
            list_ += j
    return sorted(set(list_locations + list_))


def simulation_dates_parser(date_list: list) -> list:
    """
    Parse a range of dates and returns a list of date strings.

    Examples:
        2021-01-02--2021-01-10 -> [2021-01-02 00:00:00, 2021-01-02 00:00:00, ..., 2021-01-10 00:00:00]
        2021-01-02, 2021-01-10 -> [2021-01-02 00:00:00, 2021-01-10 00:00:00]
        2021-01-02 -> [2021-01-02 00:00:00,]
    """
    dates = []
    for date_string in date_list:
        if ',' in date_string:
            dates += [
                date.strip() + ' 00:00:00' for date in date_string.split(',')
            ]
        elif '--' in date_string:
            date_start, date_end = list(
                map(lambda date: datetime.datetime.strptime(date, '%Y-%m-%d'),
                    date_string.split('--')))
            dates += [
                date.strftime('%Y-%m-%d 00:00:00') for date in [
                    date_start + datetime.timedelta(days=x)
                    for x in range(0, (date_end - date_start).days + 1)
                ]
            ]
        else:
            dates += [date_string.strip() + ' 00:00:00']

    return orm.List(list=dates)


def test_run(flexpart_code):
    """Run workflow."""

    simulation_dates = simulation_dates_parser(['2020-10-01'])
    model = ['cosmo7']
    model_offline = []
    username = 'lfernand'
    outgrid_main = 'Europe'
    outgrid_nest = 'Switzerland'
    integration_time = 24
    integration_time_offline = 0

    users_address = f'/users/{username}/resources/flexpart/'
    scratch_address = f'/scratch/snx3000/{username}/FLEXPART_input/'

    #list of locations and/or groups of locations
    list_locations = ['KIT_200magl']

    # Links to the remote files/folders.
    glc = orm.RemoteData(remote_path=users_address + 'GLC2000',
                         computer=flexpart_code.computer)
    glc_ifs = orm.RemoteData(remote_path=users_address + 'IGBP_int1.dat',
                             computer=flexpart_code.computer)
    species = orm.RemoteData(remote_path=users_address + 'SPECIES',
                             computer=flexpart_code.computer)
    surfdata = orm.RemoteData(remote_path=users_address + 'surfdata.t',
                              computer=flexpart_code.computer)
    surfdepo = orm.RemoteData(remote_path=users_address + 'surfdepo.t',
                              computer=flexpart_code.computer)

    # parent_folder = orm.load_node(pk previous tsk)
    # parent_folder = orm.RemoteData(
    #     remote_path = '/scratch/snx3000/lfernand/aiida/76/8d/cb2c-2fc6-46c4-b609-1d33fce0f60c',
    #     computer=flexpart_code.computer)
    parent_folder = None

    #builder starts
    workflow = plugins.WorkflowFactory('flexpart.multi_dates')
    builder = workflow.get_builder()
    builder.fcosmo_code = flexpart_code
    builder.fifs_code = orm.load_code('flexpart_ifs@daint')
    builder.check_meteo_ifs_code = orm.load_code(
        'check-ifs-data@daint-direct-106')
    builder.check_meteo_cosmo_code = orm.load_code(
        'check-cosmo-data@daint-direct-106')
    builder.post_processing_code = orm.load_code('post-processing@daint')

    #basic settings
    builder.simulation_dates = simulation_dates
    builder.integration_time = orm.Int(integration_time)
    builder.offline_integration_time = orm.Int(integration_time_offline)

    #meteo realted settings
    builder.model = orm.List(model)
    builder.model_offline = orm.List(model_offline)

    meteo_path = orm.List([scratch_address + mod for mod in model])
    builder.meteo_path = meteo_path
    builder.meteo_inputs = orm.Dict(
        dict=read_yaml_data('inputs/meteo_inputs.yaml', names=[
            model[-1],
        ])[model[-1]])

    if model_offline:
        meteo_path_offline = orm.List(
            [scratch_address + mod for mod in model_offline])
        builder.meteo_path_offline = meteo_path_offline
        builder.meteo_inputs_offline = orm.Dict(dict=read_yaml_data(
            'inputs/meteo_inputs.yaml', names=[
                model_offline[-1],
            ])[model_offline[-1]])

    builder.gribdir = orm.Str(scratch_address)

    #model settings
    builder.command = orm.Dict(dict=read_yaml_data(
        'inputs/command.yaml'))  #simulation date will be overwritten
    builder.input_phy = orm.Dict(dict=read_yaml_data('inputs/input_phy.yaml'))

    dict_ = read_yaml_data('inputs/locations.yaml',
                           names=make_locations_list(list_locations))
    reformated_dict_locations = reformat_locations(dict_, model[-1])
    builder.locations = orm.Dict(dict=reformated_dict_locations)

    builder.release_settings = orm.Dict(
        dict=read_yaml_data('inputs/release.yaml'))

    #other
    builder.outgrid = orm.Dict(
        dict=read_yaml_data('inputs/outgrid.yaml', names=[
            outgrid_main,
        ])[outgrid_main])
    builder.outgrid_nest = orm.Dict(
        dict=read_yaml_data('inputs/outgrid.yaml', names=[
            outgrid_nest,
        ])[outgrid_nest])
    builder.species = species
    builder.land_use = {
        'glc': glc,
        'surfdata': surfdata,
        'surfdepo': surfdepo,
    }
    builder.land_use_ifs = {
        'glc': glc_ifs,
        'surfdata': surfdata,
        'surfdepo': surfdepo,
    }
    builder.parent_calc_folder = parent_folder

    builder.flexpartcosmo.metadata.options.stash = {
        'source_list':
        ['aiida.out', 'header*', 'partposit_inst', 'grid_time_*.nc'],
        'target_base': f'/store/empa/em05/{username}/aiida_stash',
        'stash_mode': common.StashMode.COPY.value,
    }
    builder.flexpartifs.metadata.options.stash = {
        'source_list':
        ['aiida.out', 'header*', 'partposit_inst*', 'grid_time_*.nc'],
        'target_base':
        f'/store/empa/em05/{username}/aiida_stash',
        'stash_mode':
        common.StashMode.COPY.value,
    }
    builder.flexpartpost.metadata.options.stash = {
        'source_list':
        ['aiida.out', 'boundary_sensitivity_*.nc', 'grid_time_*.nc'],
        'target_base': f'/store/empa/em05/{username}/aiida_stash',
        'stash_mode': common.StashMode.COPY.value,
    }

    prepend_text_ = """#SBATCH --partition=low
    #SBATCH --account=s1152
    #SBATCH --constraint=mc
    export OMP_NUM_THREADS=${SLURM_CPUS_PER_TASK:-1}
    source $MODULESHOME/init/bash
    ulimit -s unlimited
    """

    #change wall time for cosmo and ifs in seconds
    builder.flexpartcosmo.metadata.options.max_wallclock_seconds = 1800
    #builder.flexpartifs.metadata.options.max_wallclock_seconds = 2700

    builder.flexpartcosmo.metadata.options.prepend_text = prepend_text_
    builder.flexpartifs.metadata.options.prepend_text = prepend_text_

    engine.run(builder)


@click.command()
@cmdline.utils.decorators.with_dbenv()
@cmdline.params.options.CODE()
def cli(code):
    """Run example.

    Example usage: $ ./example_01.py --code diff@localhost

    Alternative (creates diff@localhost-test code): $ ./example_01.py

    Help: $ ./example_01.py --help
    """
    test_run(code)


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



from aiida.engine import WorkChain, calcfunction
from aiida.plugins import DataFactory
from aiida import orm
from pathlib import Path
import tempfile
from netCDF4 import Dataset

NetCDF = DataFactory("netcdf.data")


def check(nc_file, version):
    """
    Checks if there is a netcdf file stored with the same name,
    if so, it checks the created date, if that is a match then returns
    False.
    """
    qb = orm.QueryBuilder()
    qb.append(
        NetCDF,
        project=[f"attributes.global_attributes.{version}"],
        filters={"attributes.filename": nc_file.attributes["filename"]},
    )
    if qb.all():
        for i in qb.all():
            if i[0] == nc_file.attributes["global_attributes"][version]:
                return False
    return True

@calcfunction
def store(remote_dir, file, time_label, nc_type):
    with tempfile.TemporaryDirectory() as td:
        remote_path = Path(remote_dir.get_remote_path()) / file.value
        temp_path = Path(td) / file.value
        remote_dir.getfile(remote_path, temp_path)

        # fill global attributes and dimensions
        nc_file = Dataset(str(temp_path), mode="r")
        nc_dimensions = {i: len(nc_file.dimensions[i]) for i in nc_file.dimensions}
        global_att = {}
        for a in nc_file.ncattrs():
            global_att[a] = repr(nc_file.getncattr(a))

        node = NetCDF(
            str(temp_path),
            remote_path=str(remote_path),
            computer=remote_dir.computer,
            g_att=global_att,
            nc_dimensions=nc_dimensions,
            other = {
                    'time_label' : time_label.value,
                    'nc_type' : nc_type.value
                    }
        )

        if "history" in node.attributes["global_attributes"].keys():
            if check(node, "history"):
                return node
        elif "created" in node.attributes["global_attributes"].keys():
            if check(node, "created"):
                return node
        else:
            return
        



class InspectWorkflow(WorkChain):
    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input_namespace("remotes", valid_type=orm.RemoteData, required=False)
        spec.input_namespace(
            "remotes_cs", valid_type=orm.RemoteStashFolderData, required=False
        )
        spec.input('time_label', valid_type=orm.Str, required=False)
        spec.input('nc_type', valid_type=orm.Str, required=False)
        spec.outputs.dynamic = True
        spec.outline(
            cls.fill_remote_data,
            cls.inspect,
        )
    def fill_remote_data(self):
        self.ctx.dict_remote_data = {}
        if "remotes" in self.inputs:
            self.ctx.dict_remote_data = self.inputs.remotes
        else:
            for k, v in self.inputs.remotes_cs.items():
                self.ctx.dict_remote_data[k] = orm.RemoteData(
                    remote_path=v.target_basepath, computer=v.computer
                )

    def inspect(self):
        for _, i in self.ctx.dict_remote_data.items():
            for file in i.listdir():
                if ".nc" in file:
                    store(i, file, self.inputs.time_label, self.inputs.nc_type)



# -*- coding: utf-8 -*-
"""Flexpart multi-dates WorkChain."""
from aiida import engine, plugins, orm
from aiida_flexpart.workflows.child_meteo_workflow import TransferMeteoWorkflow

# plugins
FlexpartCosmoCalculation = plugins.CalculationFactory("flexpart.cosmo")
FlexpartIfsCalculation = plugins.CalculationFactory("flexpart.ifs")
FlexpartPostCalculation = plugins.CalculationFactory("flexpart.post")

# possible models
cosmo_models = ["cosmo7", "cosmo1", "kenda1"]
ECMWF_models = ["IFS_GL_05", "IFS_GL_1", "IFS_EU_02", "IFS_EU_01"]


class FlexpartSimWorkflow(engine.WorkChain):
    """Flexpart multi-dates workflow"""

    @classmethod
    def define(cls, spec):
        """Specify inputs and outputs."""
        super().define(spec)

        # Codes
        spec.input("fcosmo_code", valid_type=orm.AbstractCode)
        spec.input("fifs_code", valid_type=orm.AbstractCode)
        spec.input("post_processing_code", valid_type=orm.AbstractCode)

        # Basic Inputs
        spec.expose_inputs(TransferMeteoWorkflow)
        spec.input("date", valid_type=orm.Str, required=False)

        # Model settings
        # Command is exposed form TransferMeteoWorkflow
        spec.input("input_phy", valid_type=orm.Dict)
        spec.input("release_settings", valid_type=orm.Dict)
        spec.input("locations", valid_type=orm.Dict, 
                   help="Dictionary of locations properties."
        )

        # Meteo related inputs
        spec.input(
            "meteo_inputs",
            valid_type=orm.Dict,
            required=False,
            help="Meteo models input params.",
        )
        spec.input(
            "meteo_inputs_offline",
            valid_type=orm.Dict,
            required=False,
            help="Meteo models input params.",
        )
        spec.input(
            "meteo_path",
            valid_type=orm.List,
            required=False,
            help="Path to the folder containing the meteorological input data.",
        )
        spec.input(
            "meteo_path_offline",
            valid_type=orm.List,
            required=False,
            help="Path to the folder containing the meteorological input data.",
        )

        # Others
        spec.input("outgrid", valid_type=orm.Dict)
        spec.input("outgrid_nest", valid_type=orm.Dict, required=False)
        spec.input("species", valid_type=orm.RemoteData, required=True)
        spec.input_namespace(
            "land_use",
            valid_type=orm.RemoteData,
            required=False,
            dynamic=True,
            help="#TODO",
        )
        spec.input_namespace(
            "land_use_ifs", valid_type=orm.RemoteData, required=False, dynamic=True
        )
        spec.expose_inputs(
            FlexpartCosmoCalculation,
            include=["metadata.options"],
            namespace="flexpartcosmo",
        )
        spec.expose_inputs(
            FlexpartIfsCalculation,
            include=["metadata.options"],
            namespace="flexpartifs",
        )
        spec.expose_inputs(
            FlexpartPostCalculation,
            include=["metadata.options"],
            namespace="flexpartpost",
        )
        spec.outputs.dynamic = True

        #exit codes
        spec.exit_code(400, 'ERROR_CALCULATION_FAILED', 
                       'the previous calculation did not finish successfully')

        spec.outline(
            cls.setup,

            engine.if_(cls.run_cosmo)(
                cls.run_cosmo_simulation,
                cls.inspect_calculation
                ),
            engine.if_(cls.run_ifs)(
                cls.run_ifs_simulation,
                cls.inspect_calculation
                ),

            cls.post_processing,
            cls.results,
        )

    def run_cosmo(self):
        """run cosmo simulation"""
        if all(mod in cosmo_models for mod in self.inputs.model) and self.inputs.model:
            return True
        return False

    def run_ifs(self):
        """run ifs simulation"""
        if (
            all(mod in ECMWF_models for mod in self.inputs.model)
            or all(mod in ECMWF_models for mod in self.inputs.model_offline)
            and self.inputs.model
            and self.inputs.model_offline
        ):
            return True
        return False
    
    def inspect_calculation(self):
        if not self.ctx.calculations[-1].is_finished_ok:
            self.report('ERROR calculation did not finish ok')
            return self.exit_codes.ERROR_CALCULATION_FAILED
        self.report('calculation successfull')

    def setup(self):

        self.ctx.simulation_date = self.inputs.date.value
        self.ctx.integration_time = self.inputs.integration_time
        self.ctx.offline_integration_time = self.inputs.offline_integration_time

        # model settings
        self.ctx.release_settings = self.inputs.release_settings
        self.ctx.command = self.inputs.command
        self.ctx.input_phy = self.inputs.input_phy
        self.ctx.locations = self.inputs.locations

        # others
        self.ctx.outgrid = self.inputs.outgrid
        self.ctx.species = self.inputs.species
        self.ctx.land_use = self.inputs.land_use

    def post_processing(self):
        """post processing"""
        self.report("starting post-processsing")
        builder = FlexpartPostCalculation.get_builder()
        builder.code = self.inputs.post_processing_code
        builder.input_dir = self.ctx.calculations[-1].outputs.remote_folder

        if self.ctx.offline_integration_time > 0:

            builder.input_dir = self.ctx.calculations[-2].outputs.remote_folder
            builder.input_offline_dir = self.ctx.calculations[-1].outputs.remote_folder

        builder.metadata.options = self.inputs.flexpartpost.metadata.options

        running = self.submit(builder)
        self.to_context(calculations=engine.append_(running))

    def run_cosmo_simulation(self):
        """Run calculations for equation of state."""

        self.report(f"starting flexpart cosmo {self.ctx.simulation_date}")

        builder = FlexpartCosmoCalculation.get_builder()
        builder.code = self.inputs.fcosmo_code

        # update command file
        new_dict = self.ctx.command.get_dict()
        new_dict["simulation_date"] = self.ctx.simulation_date
        new_dict["age_class"] = self.inputs.integration_time * 3600
        new_dict.update(self.inputs.meteo_inputs)

        # model settings
        builder.model_settings = {
            "release_settings": self.ctx.release_settings,
            "locations": self.ctx.locations,
            "command": orm.Dict(dict=new_dict),
            "input_phy": self.ctx.input_phy,
        }

        builder.outgrid = orm.Dict(
            list(self.ctx.outgrid.get_dict().values())[0])
        if 'outgrid_nest' in self.inputs:
            builder.outgrid_nest = orm.Dict(
                list(self.inputs.outgrid_nest.get_dict().values())[0])
        builder.species = self.ctx.species
        builder.land_use = self.ctx.land_use
        builder.meteo_path = self.inputs.meteo_path

        # Walltime, memory, and resources.
        builder.metadata.description = "Test workflow to submit a flexpart calculation"
        builder.metadata.options = self.inputs.flexpartcosmo.metadata.options

        # Ask the workflow to continue when the results are ready and store them in the context
        running = self.submit(builder)
        self.to_context(calculations=engine.append_(running))

    def run_ifs_simulation(self):
        """Run calculations for equation of state."""
        # Set up calculation.
        self.report(f"running flexpart ifs for {self.ctx.simulation_date}")
        builder = FlexpartIfsCalculation.get_builder()
        builder.code = self.inputs.fifs_code

        # changes in the command file
        new_dict = self.ctx.command.get_dict()
        new_dict["simulation_date"] = self.ctx.simulation_date

        if self.ctx.offline_integration_time > 0:
            new_dict["age_class"] = self.ctx.offline_integration_time * 3600
            new_dict["dumped_particle_data"] = True

            self.ctx.parent_calc_folder = self.ctx.calculations[
                -1
            ].outputs.remote_folder
            builder.parent_calc_folder = self.ctx.parent_calc_folder
            self.report(f"starting from: {self.ctx.parent_calc_folder}")

            builder.meteo_path = self.inputs.meteo_path_offline
            new_dict.update(self.inputs.meteo_inputs_offline)

        else:
            new_dict["age_class"] = self.inputs.integration_time * 3600
            builder.meteo_path = self.inputs.meteo_path
            new_dict.update(self.inputs.meteo_inputs)

        # model settings
        builder.model_settings = {
            "release_settings": self.ctx.release_settings,
            "locations": self.ctx.locations,
            "command": orm.Dict(dict=new_dict),
        }

        builder.outgrid = orm.Dict(
            list(self.ctx.outgrid.get_dict().values())[0])
        if 'outgrid_nest' in self.inputs:
            builder.outgrid_nest = orm.Dict(
                list(self.inputs.outgrid_nest.get_dict().values())[0])
        builder.species = self.ctx.species
        builder.land_use = self.inputs.land_use_ifs

        # Walltime, memory, and resources.
        builder.metadata.description = "Test workflow to submit a flexpart calculation"
        builder.metadata.options = self.inputs.flexpartifs.metadata.options

        # Ask the workflow to continue when the results are ready and store them in the context
        running = self.submit(builder)
        self.to_context(calculations=engine.append_(running))

    def results(self):
        """Process results."""
        for indx, calculation in enumerate(self.ctx.calculations):
            self.out(f"calculation_{indx}_output_file", calculation.outputs.output_file)



import os
from aiida.orm import RemoteData


class NetCdfData(RemoteData):

    def __init__(
        self, 
        filepath=None,
        remote_path=None,
        g_att=None, 
        nc_dimensions=None,
        other = {},
        **kwargs,
    ):
        """
        Data plugin for Netcdf files.
        """
        super().__init__(**kwargs)
        if filepath is not None:
            filename = os.path.basename(filepath)
            self.set_remote_path(remote_path)
            self.set_filename(filename)
            self.set_global_attributes(g_att, nc_dimensions)
        if other:
            for k,v in other.items():
                self.base.attributes.set(k, v)


    def set_filename(self, val):
        self.base.attributes.set("filename", val)

    def set_global_attributes(self, g_att, nc_dimensions):
        self.base.attributes.set("global_attributes", g_att)
        self.base.attributes.set("dimensions", nc_dimensions)

    def ncdump(self):
        """Small python version of ncdump."""
        print("dimensions:")
        for k, v in self.base.attributes.get("dimensions").items():
            print(f"\t {k} = {v}")
        print("// global attributes:")
        for k, v in self.base.attributes.get("global_attributes").items():
            print(f"\t :{k} = {v}")



# -*- coding: utf-8 -*-
from aiida import engine, orm
from aiida_flexpart.workflows.child_meteo_workflow import TransferMeteoWorkflow
from aiida_flexpart.workflows.child_sim_workflow import FlexpartSimWorkflow



class ParentWorkflow(engine.WorkChain):
    @classmethod
    def define(cls, spec):
        super().define(spec)

        #extras
        spec.input('name', valid_type=str, non_db=True, required=False)

        spec.expose_inputs(TransferMeteoWorkflow)
        spec.expose_outputs(TransferMeteoWorkflow)
        spec.expose_inputs(FlexpartSimWorkflow)
        spec.expose_outputs(FlexpartSimWorkflow)
        spec.outputs.dynamic = True

        spec.outline(
            cls.setup,
            cls.transfer_meteo,
            cls.run_sim,
            cls.finalize,
        )

    def setup(self):
        #self.inputs.simulation_dates is orm.List
        self.ctx.month_chunks = 1
        if 'name' in self.inputs:
            out_n = 'None'
            if 'outgrid_nest' in self.inputs:
                out_n = self.inputs.outgrid_nest.get_dict()
            self.node.base.extras.set(
                self.inputs.name, {
                    'command': self.inputs.command.get_dict(),
                    'input_phy': self.inputs.input_phy.get_dict(),
                    'release': self.inputs.release_settings.get_dict(),
                    'locations': self.inputs.locations.get_dict(),
                    'integration_time': self.inputs.integration_time.value,
                    'offline_integration_time':
                    self.inputs.offline_integration_time.value,
                    'model': self.inputs.model,
                    'model_offline': self.inputs.model_offline,
                    'outgrid': self.inputs.outgrid.get_dict(),
                    'outgrid_nest': out_n
                })

    def transfer_meteo(self):
        child_1 = self.submit(TransferMeteoWorkflow,
                              **self.exposed_inputs(TransferMeteoWorkflow))
        return engine.ToContext(child_1=child_1)

    def run_sim(self):

        for i in range(len(self.inputs.simulation_dates)):
            child = self.submit(
                FlexpartSimWorkflow,
                **self.exposed_inputs(FlexpartSimWorkflow),
                date=orm.Str(self.inputs.simulation_dates[i]),
            )
            self.to_context(workchains=engine.append_(child))

    def finalize(self):
        self.out_many(
            self.exposed_outputs(self.ctx.child_1, TransferMeteoWorkflow))
        for w in self.ctx.workchains:
            self.out_many(self.exposed_outputs(w, FlexpartSimWorkflow))
